

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `NPLCSkeleton'                                                //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class NPLCSkeleton {
public :
  NPLCSkeleton(const Frame* _loc, const Frame* _rem,
               const Frame** _subs, int** subsData,
               const Frame* _msk, int* mskData,
	       Data_Types _data_type_input,
	       Operator_Types _operator_type_input) ;

  ~NPLCSkeleton() ;

  void execute(void* source_array, void* target_array) ;

protected :
  const Frame* loc ;
  const Frame* rem ;
  const Frame** subs ;
  const Frame*  msk ;

  Stride* tar_strs ;
  int MP ;

private :
  int rank ;  // of `loc'

  void loop(int locOffset, int** subsElem, int* mskElem, const int r) ;

  int *apg2ini, *ini2apg ;

  MPI_Group apg_group ;
  MPI_Comm apg_comm ;

  int NP, s_N_all_input ;
  int *source_lidx, *index_lidx, *index_pnum ;

  Data_Types data_type_input ;
  Operator_Types operator_type_input ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class `ScatterOp': deals with transfer between                     //
//                     ``immediate target'' as seen by Florida         //
//                     routine, and Adlib/HPF-level target array.      //
//                                                                     //
//                     Incorporates broadcast if target has            //
//                     replicated alignment                            //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
class ScatterOp : public NPLCSkeleton {
public :
  ScatterOp(const Frame* loc, const Frame* rem, const int len,
	    const Frame** subs, int** subsData,
	    const Frame* msk, int* mskData,
	    Data_Types _data_type_input,
	    Operator_Types _operator_type_input) :
      NPLCSkeleton(loc, rem, subs, subsData, msk, mskData,
	           _data_type_input, _operator_type_input),
      rank(rem->rnk()), bcst(rem, len) {
    target_array = new T [MP] ;

    exts     = new int [rank] ;
    dstSteps = new int [rank] ;
    srcSteps = new int [rank] ;
  }

  ~ScatterOp() {
    delete [] srcSteps ;
    delete [] dstSteps ;
    delete [] exts ;

    delete [] target_array ;
  }

  void execute(T* _locBase, T* _remBase) {
    opUnit() ;  // Set all elements of `target_array' to unit

    NPLCSkeleton :: execute(_locBase, target_array) ;

    // Accumulate active elements of `target_array' into `rem'...

    remBase = _remBase ;
    if(rem->grp().member())
      loop(0, 0, 0) ;

    // Broadcast elements of `rem' if it has replicated alignment...

    bcst.execute(remBase) ;
  }

protected :
  Replicate bcst ;

  T *target_array, *remBase ;

  int *exts, *srcSteps, *dstSteps ;

  int rank ;  // of `rem'

  virtual void opUnit() = 0 ;
  virtual void opMap(const int dstOffset, const int srcOffset) = 0 ;

private :

  void loop(int dstOffset, int srcOffset, const int r) {

    // Multi-dimensional loop to accumulate target values into `rem'.

    if(r == rank)
      opMap(dstOffset, srcOffset) ;
    else {
      Stride dstStr = rem->str(r) ;
      Stride srcStr = tar_strs [r] ;

      LocBlocksIndex i(rem->rng(r)) ;
      for(i.beginLocBlk() ; i.test() ; i.nextLocBlk()) {
        exts     [r] = i.count ;
        dstSteps [r] = dstStr.step(i) ;
        srcSteps [r] = srcStr.step(i) ;

        loop(dstOffset + dstStr.offset(i), srcOffset + srcStr.offset(i),
             r + 1) ;
      }
    }
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  block ``accumulation'' functions.                                  //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
inline void blockSum(T* res, int* resStrs, T* src, int* srcStrs,
                     int* exts, int rank) {
  if(rank == 0)
    *res += *src ;
  else if(rank == 1) {
    int count = *exts ;
    int srcStep = *srcStrs ;
    int resStep = *resStrs ;
    for(int i = 0 ; i < count ; i++) {
      *res += *src ;
      src  += srcStep ;
      res  += resStep ;
    }
  }
  else {
    int count = *exts ;
    int srcStep = *srcStrs ;
    int resStep = *resStrs ;
    rank-- ;
    exts++ ;
    resStrs++ ;
    srcStrs++ ;
    for(int j = 0 ; j < count ; j++) {
      blockSum(res, resStrs, src, srcStrs, exts, rank) ;
      res += resStep ;
      src += srcStep ;
    }
  }
} ;

template<class T>
inline void blockProduct(T* res, int* resStrs, T* src, int* srcStrs,
                         int* exts, int rank) {
  if(rank == 0)
    *res *= *src ;
  else if(rank == 1) {
    int count = *exts ;
    int srcStep = *srcStrs ;
    int resStep = *resStrs ;
    for(int i = 0 ; i < count ; i++) {
      *res *= *src ;
      src  += srcStep ;
      res  += resStep ;
    }
  }
  else {
    int count = *exts ;
    int srcStep = *srcStrs ;
    int resStep = *resStrs ;
    rank-- ;
    exts++ ;
    resStrs++ ;
    srcStrs++ ;
    for(int j = 0 ; j < count ; j++) {
      blockProduct(res, resStrs, src, srcStrs, exts, rank) ;
      res += resStep ;
      src += srcStep ;
    }
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Compute data types                                                 //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

inline Data_Types type_of(int* dummy) {
  return TYPE_int ;
}

inline Data_Types type_of(float* dummy) {
  return TYPE_float ;
}

inline Data_Types type_of(double* dummy) {
  return TYPE_double ;
}

inline Data_Types type_of(COMPLEX* dummy) {
  return TYPE_complex ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class `ScatterSum'                                                 //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
class ScatterSum : public ScatterOp<T> {
public :
  ScatterSum(const Frame* loc, const Frame* rem, const int len,
	     const Frame** subs, int** subsData,
	     const Frame* msk, int* mskData) :
      ScatterOp<T>(loc, rem, len, subs, subsData, msk, mskData,
                   type_of(target_array), SUM) {}

private :

  // Virtual functions for arithmetic on ``immediate target'' array.

  void opUnit() {
    for(int i = 0 ; i < MP ; i++)
      target_array [i] = 0 ;
  }

  void opMap(const int dstOffset, const int srcOffset) {
    blockSum(remBase + dstOffset,      dstSteps,
             target_array + srcOffset, srcSteps,
             exts, rank) ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class `ScatterProduct'                                             //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
class ScatterProduct : public ScatterOp<T> {
public :
  ScatterProduct(const Frame* loc, const Frame* rem, const int len,
	     const Frame** subs, int** subsData,
	     const Frame* msk, int* mskData) :
      ScatterOp<T>(loc, rem, len, subs, subsData, msk, mskData,
                   type_of(target_array), PRODUCT) {}

private :

  // Virtual functions for arithmetic on ``immediate target'' array.

  void opUnit() {
    for(int i = 0 ; i < MP ; i++)
      target_array [i] = 1 ;
  }

  void opMap(const int dstOffset, const int srcOffset) {
    blockProduct(remBase + dstOffset,      dstSteps,
             target_array + srcOffset, srcSteps,
             exts, rank) ;
  }
} ;


